@@ -4,12 +4,12 @@ import logging
 from typing import Dict, List, Optional, Tuple, Union
 
 from expertsystem.particle import ParticleCollection, Spin
+from expertsystem.reaction import Result
 from expertsystem.reaction.combinatorics import (
     perform_external_edge_identical_particle_combinatorics,
 )
 from expertsystem.reaction.quantum_numbers import ParticleWithSpin
-from expertsystem.reaction.solving import Result
-from expertsystem.reaction.topology import StateTransitionGraph, Topology
+from expertsystem.reaction.topology import StateTransitionGraph
 
 from .model import (
     AmplitudeModel,
@@ -25,11 +25,9 @@ from .model import (
     IncoherentIntensity,
     IntensityNode,
     Kinematics,
-    NormalizedIntensity,
     ParticleDynamics,
     RecoilSystem,
     SequentialAmplitude,
-    StrengthIntensity,
 )
 
 
@@ -47,13 +45,16 @@ def _group_graphs_same_initial_and_final(
         Tuple[tuple, tuple], List[StateTransitionGraph[ParticleWithSpin]]
     ] = dict()
     for graph in graphs:
-        ise = graph.get_final_state_edges()
-        fse = graph.get_initial_state_edges()
+        ise = graph.get_final_state_edge_ids()
+        fse = graph.get_initial_state_edge_ids()
         graph_group = (
             tuple(
                 sorted(
                     [
-                        (graph.edge_props[x][0].name, graph.edge_props[x][1])
+                        (
+                            graph.get_edge_props(x)[0].name,
+                            graph.get_edge_props(x)[1],
+                        )
                         for x in ise
                     ]
                 )
@@ -61,7 +62,10 @@ def _group_graphs_same_initial_and_final(
             tuple(
                 sorted(
                     [
-                        (graph.edge_props[x][0].name, graph.edge_props[x][1])
+                        (
+                            graph.get_edge_props(x)[0].name,
+                            graph.get_edge_props(x)[1],
+                        )
                         for x in fse
                     ]
                 )
@@ -80,8 +84,8 @@ def _get_graph_group_unique_label(
 ) -> str:
     label = ""
     if graph_group:
-        ise = graph_group[0].get_initial_state_edges()
-        fse = graph_group[0].get_final_state_edges()
+        ise = graph_group[0].get_initial_state_edge_ids()
+        fse = graph_group[0].get_final_state_edge_ids()
         is_names = _get_name_hel_list(graph_group[0], ise)
         fs_names = _get_name_hel_list(graph_group[0], fse)
         label += (
@@ -93,7 +97,7 @@ def _get_graph_group_unique_label(
 
 
 def _determine_attached_final_state(
-    graph: Topology, edge_id: int
+    graph: StateTransitionGraph[ParticleWithSpin], edge_id: int
 ) -> List[int]:
     """Determine all final state particles of a graph.
 
@@ -101,7 +105,7 @@ def _determine_attached_final_state(
     the root).
     """
     final_state_edge_ids = []
-    all_final_state_edges = graph.get_final_state_edges()
+    all_final_state_edges = graph.get_final_state_edge_ids()
     current_edges = [edge_id]
     while current_edges:
         temp_current_edges = current_edges
@@ -111,18 +115,21 @@ def _determine_attached_final_state(
                 final_state_edge_ids.append(current_edge)
             else:
                 node_id = graph.edges[current_edge].ending_node_id
-                current_edges.extend(
-                    graph.get_edges_outgoing_from_node(node_id)
-                )
+                if node_id:
+                    current_edges.extend(
+                        graph.get_edge_ids_outgoing_from_node(node_id)
+                    )
     return final_state_edge_ids
 
 
-def _get_recoil_edge(graph: Topology, edge_id: int) -> Optional[int]:
+def _get_recoil_edge(
+    graph: StateTransitionGraph[ParticleWithSpin], edge_id: int
+) -> Optional[int]:
     """Determine the id of the recoil edge for the specified edge of a graph."""
     node_id = graph.edges[edge_id].originating_node_id
     if node_id is None:
         return None
-    outgoing_edges = graph.get_edges_outgoing_from_node(node_id)
+    outgoing_edges = graph.get_edge_ids_outgoing_from_node(node_id)
     outgoing_edges.remove(edge_id)
     if len(outgoing_edges) != 1:
         raise ValueError(
@@ -132,12 +139,14 @@ def _get_recoil_edge(graph: Topology, edge_id: int) -> Optional[int]:
     return outgoing_edges[0]
 
 
-def _get_parent_recoil_edge(graph: Topology, edge_id: int) -> Optional[int]:
+def _get_parent_recoil_edge(
+    graph: StateTransitionGraph[ParticleWithSpin], edge_id: int
+) -> Optional[int]:
     """Determine the id of the recoil edge of the parent edge."""
     node_id = graph.edges[edge_id].originating_node_id
     if node_id is None:
         return None
-    ingoing_edges = graph.get_edges_ingoing_to_node(node_id)
+    ingoing_edges = graph.get_edge_ids_ingoing_to_node(node_id)
     if len(ingoing_edges) != 1:
         raise ValueError(
             f"The node with id {node_id} does not have a single ingoing edge!\n"
@@ -148,22 +157,15 @@ def _get_parent_recoil_edge(graph: Topology, edge_id: int) -> Optional[int]:
 
 def _get_prefactor(
     graph: StateTransitionGraph[ParticleWithSpin],
-) -> Optional[float]:
+) -> float:
     """Calculate the product of all prefactors defined in this graph."""
-    prefactor = None
+    prefactor = 1.0
     for node_id in graph.nodes:
-        if node_id in graph.node_props:
-            temp_prefactor = __validate_float_type(
-                graph.node_props[node_id].parity_prefactor
-            )
+        node_props = graph.get_node_props(node_id)
+        if node_props:
+            temp_prefactor = __validate_float_type(node_props.parity_prefactor)
             if temp_prefactor is not None:
-                if prefactor is None:
-                    prefactor = temp_prefactor
-                else:
-                    prefactor *= temp_prefactor
-            else:
-                prefactor = None
-                break
+                prefactor *= temp_prefactor
     return prefactor
 
 
@@ -172,7 +174,7 @@ def _generate_particle_collection(
 ) -> ParticleCollection:
     particles = ParticleCollection()
     for graph in graphs:
-        for edge_props in graph.edge_props.values():
+        for edge_props in map(graph.get_edge_props, graph.edges):
             particle_name = edge_props[0].name
             if particle_name not in particles:
                 particles.add(edge_props[0])
@@ -189,7 +191,7 @@ def _generate_kinematics(
         initial_state=initial_state,
         final_state=final_state,
         intermediate_states=len(
-            result.solutions[0].get_intermediate_state_edges()
+            result.solutions[0].get_intermediate_state_edge_ids()
         ),
     )
     return kinematics
@@ -217,11 +219,11 @@ def _get_name_hel_list(
 ) -> List[Tuple[str, float]]:
     name_hel_list = []
     for i in edge_ids:
-        temp_hel = graph.edge_props[i][1]
+        temp_hel = graph.get_edge_props(i)[1]
         # remove .0
         if temp_hel % 1 == 0:
             temp_hel = int(temp_hel)
-        name_hel_list.append((graph.edge_props[i][0].name, temp_hel))
+        name_hel_list.append((graph.get_edge_props(i)[0].name, temp_hel))
 
     # in order to ensure correct naming of amplitude coefficients the list has
     # to be sorted by name. The same coefficient names have to be created for
@@ -272,6 +274,9 @@ class _HelicityAmplitudeNameGenerator:
                 priority_partner_coefficient_suffix,
             ) = self._generate_amplitude_coefficient_couple(graph, node_id)
 
+            if graph.get_node_props(node_id).parity_prefactor is None:
+                continue
+
             if (
                 coefficient_suffix
                 not in self.parity_partner_coefficient_mapping
@@ -315,7 +320,7 @@ class _HelicityAmplitudeNameGenerator:
         """
         name = ""
         if isinstance(node_id, int):
-            nodelist = {node_id}
+            nodelist = frozenset({node_id})
         else:
             nodelist = graph.nodes
         for node in nodelist:
@@ -335,8 +340,8 @@ class _HelicityAmplitudeNameGenerator:
     def _retrieve_helicity_info(
         graph: StateTransitionGraph[ParticleWithSpin], node_id: int
     ) -> Tuple[List[Tuple[str, float]], List[Tuple[str, float]]]:
-        in_edges = graph.get_edges_ingoing_to_node(node_id)
-        out_edges = graph.get_edges_outgoing_from_node(node_id)
+        in_edges = graph.get_edge_ids_ingoing_to_node(node_id)
+        out_edges = graph.get_edge_ids_outgoing_from_node(node_id)
 
         in_names_hel_list = _get_name_hel_list(graph, in_edges)
         out_names_hel_list = _get_name_hel_list(graph, out_edges)
@@ -362,17 +367,10 @@ class _HelicityAmplitudeNameGenerator:
         """Generate unique suffix for a sequential amplitude graph."""
         output_suffix = ""
         for node_id in graph.nodes:
-            raw_suffix = self.generate_amplitude_coefficient_name(
-                graph, node_id
-            )
-            if raw_suffix not in self.parity_partner_coefficient_mapping:
-                raise KeyError(
-                    f"Coefficient name {raw_suffix} not found in mapping!"
-                )
-            coefficient_suffix = self.parity_partner_coefficient_mapping[
-                raw_suffix
-            ]
-            output_suffix += coefficient_suffix + ";"
+            suffix = self.generate_amplitude_coefficient_name(graph, node_id)
+            if suffix in self.parity_partner_coefficient_mapping:
+                suffix = self.parity_partner_coefficient_mapping[suffix]
+            output_suffix += suffix + ";"
         return output_suffix
 
 
@@ -434,14 +432,11 @@ class HelicityAmplitudeGenerator:
                 graph_group
             )
             incoherent_intensity.intensities.append(coherent_intensity)
-        if len(incoherent_intensity.intensities) > 1:
-            intensity: IntensityNode = NormalizedIntensity(
-                incoherent_intensity
-            )
-        else:
-            intensity = incoherent_intensity.intensities[0]
-        strength_intensity = self.__prepend_strength(intensity)
-        return strength_intensity
+        if len(incoherent_intensity.intensities) == 0:
+            raise ValueError("List of incoherent intensities cannot be empty")
+        if len(incoherent_intensity.intensities) == 1:
+            return incoherent_intensity.intensities[0]
+        return incoherent_intensity
 
     def __create_parameter_couplings(
         self, graph_groups: List[List[StateTransitionGraph[ParticleWithSpin]]]
@@ -468,20 +463,6 @@ class HelicityAmplitudeGenerator:
                 )
         return coherent_intensity
 
-    def __prepend_strength(
-        self, intensity: IntensityNode
-    ) -> StrengthIntensity:
-        strength = self.__register_parameter(
-            name="strength_incoherent",
-            value=1.0,
-            fix=True,
-        )
-        return StrengthIntensity(
-            component="incoherent_with_strength",
-            strength=strength,
-            intensity=intensity,
-        )
-
     def __generate_sequential_decay(
         self, graph: StateTransitionGraph[ParticleWithSpin]
     ) -> AmplitudeNode:
@@ -503,9 +484,7 @@ class HelicityAmplitudeGenerator:
         )
 
     def _generate_partial_decay(
-        self,
-        graph: StateTransitionGraph[ParticleWithSpin],
-        node_id: Optional[int] = None,
+        self, graph: StateTransitionGraph[ParticleWithSpin], node_id: int
     ) -> DecayNode:
         def create_helicity_particle(
             edge_props: ParticleWithSpin,
@@ -519,8 +498,8 @@ class HelicityAmplitudeGenerator:
             return HelicityParticle(particle, helicity)
 
         decay_products: List[DecayProduct] = list()
-        for out_edge_id in graph.get_edges_outgoing_from_node(node_id):
-            edge_props = graph.edge_props[out_edge_id]
+        for out_edge_id in graph.get_edge_ids_outgoing_from_node(node_id):
+            edge_props = graph.get_edge_props(out_edge_id)
             helicity_particle = create_helicity_particle(edge_props)
             final_state_ids = _determine_attached_final_state(
                 graph, out_edge_id
@@ -533,11 +512,11 @@ class HelicityAmplitudeGenerator:
                 )
             )
 
-        in_edge_ids = graph.get_edges_ingoing_to_node(node_id)
+        in_edge_ids = graph.get_edge_ids_ingoing_to_node(node_id)
         if len(in_edge_ids) != 1:
             raise ValueError("This node does not represent a two body decay!")
         ingoing_edge_id = in_edge_ids[0]
-        edge_props = graph.edge_props[ingoing_edge_id]
+        edge_props = graph.get_edge_props(ingoing_edge_id)
         helicity_particle = create_helicity_particle(edge_props)
         helicity_decay = HelicityDecay(helicity_particle, decay_products)
 
@@ -582,20 +561,24 @@ class HelicityAmplitudeGenerator:
         self, graph: StateTransitionGraph[ParticleWithSpin]
     ) -> Optional[float]:
         prefactor = _get_prefactor(graph)
-        for node_id in graph.nodes:
-            raw_suffix = (
-                self.name_generator.generate_amplitude_coefficient_name(
-                    graph, node_id
+        if prefactor != 1.0:
+            for node_id in graph.nodes:
+                raw_suffix = (
+                    self.name_generator.generate_amplitude_coefficient_name(
+                        graph, node_id
+                    )
                 )
-            )
-            coefficient_suffix = (
-                self.name_generator.parity_partner_coefficient_mapping[
+                if (
                     raw_suffix
-                ]
-            )
-            if coefficient_suffix != raw_suffix:
-                if prefactor != 1.0 and prefactor is not None:
-                    return prefactor
+                    in self.name_generator.parity_partner_coefficient_mapping
+                ):
+                    coefficient_suffix = (
+                        self.name_generator.parity_partner_coefficient_mapping[
+                            raw_suffix
+                        ]
+                    )
+                    if coefficient_suffix != raw_suffix:
+                        return prefactor
         return None
 
     def __register_parameter(
@@ -603,7 +586,7 @@ class HelicityAmplitudeGenerator:
     ) -> FitParameter:
         if name in self.fit_parameters:
             return self.fit_parameters[name]
-        parameter = FitParameter(name=name, value=value, is_fixed=fix)
+        parameter = FitParameter(name=name, value=value, fix=fix)
         self.fit_parameters.add(parameter)
         return parameter
 
